#!/usr/bin/env python

################################################################################
# PAV Pipeline: spawn jellyfish k-mer counter instances
#
# Hong Ching Lee     04 January 2012
################################################################################

# import libraries
import argparse # parse input options
import ConfigParser # parse the config file
import os
import sys
import glob
import math
import subprocess
import logging
import time
import re


def execute_command(cmd, is_verbose):
    if is_verbose:
        print cmd

    try:
        return_code = subprocess.call(cmd, shell=True)
        logging.info('Executing %s\n', cmd)

        if return_code != 0:
            print >> sys.stderr, "Problems executing command: " + cmd
    except OSError, e:
        msg = "Execution of command \"" + cmd + "\" failed: " + e
        logging.critical(msg)
        print >> sys.stderr, msg
        print >> sys.stderr, cmd


def jellyfish_bloom(config, sample, out_dir, tag, ncpus, mode, min, max, is_verbose):
    '''Run jellyfish on a range of k-mer sizes.'''

    coords = [] # list of x, y coordinates with x being k-mer size, y being the corresponding uniqueness ratio
    log_string = '\nK\tUniqueness Ratio\tUnique\tDistinct\tTotal\n'

    print "k\tuniqueness ratio"

    for i in range(min, max + 1):
        ratio, tmp_log = jellyfish_polyp(config, i, sample, out_dir, tag, ncpus, mode, is_verbose)
        coords.append((i, ratio))
        print i, "\t", ratio
        log_string = log_string + str(i) + '\t' + str(ratio) + tmp_log

    # find the knee point
    start, end = find_line_ends(coords)
    p1 = coords[start]
    p2 = coords[end]
    max_diff = 0
    max_i = start + min

    for i in range(start, end + 1):
        y = line_y(p1, p2, i + min)
        diff = abs(coords[i][1] - y)

        if diff > max_diff:
            max_diff = diff
            max_i = i + min

    suggest_string = "\nSuggested k-mer size for this sample is: " + str(max_i) + "\n"
    log_string = log_string + suggest_string
    logging.info(log_string)
    print suggest_string


def jellyfish_polyp(config, kmer_size, sample, out_dir, tag, ncpus, mode, is_verbose):
    '''Invoke one instance of jellyfish on a specific k-mer size.'''

    jellyfish = config.get('JELLYFISH', 'path') + '/jellyfish'
    output_name = out_dir + "/" + tag + "_k" + str(kmer_size)
    in_files = " ".join(sample)
    to_delete = [] # intermediate files to be cleaned up afterwards

    # Run jellyfish count
    count_cmd = jellyfish + " count --both-strands -m " + str(kmer_size) +\
        " -c " + config.get('ADVANCED', 'jellyfish_counter_bits') +\
        " -s " + config.get('ADVANCED', 'jellyfish_hash_size') +\
        " -t " + str(ncpus) + " -o " + output_name + "_count " + in_files
    execute_command(count_cmd, is_verbose)

    # Run jellyfish merge
    unmerged_db = glob.glob(output_name + "_count*")
    to_delete = list(unmerged_db)

    if len(unmerged_db) > 1:
        output_name = output_name + "_merged"
        merge_cmd = jellyfish + " merge -o " + output_name + " " + " ".join(unmerged_db)
        execute_command(merge_cmd, is_verbose)
        to_delete.append(output_name)
    else:
        output_name = output_name + "_count_0"

    # Run either jellyfish dump or stats depending on the specified mode
    if mode == 'dump':
        dump_cmd = jellyfish + " dump -c -t -o " + out_dir + "/" + tag + " " + output_name
        execute_command(dump_cmd, is_verbose)
        clean_up_files(to_delete, is_verbose)
    elif mode == 'stats':
        stats_string = subprocess.check_output([jellyfish, 'stats', output_name])
        ratio, log_string = interpret_stats(stats_string)
        clean_up_files(to_delete, is_verbose)

        return ratio, log_string


def find_line_ends(coords):
    '''Go through the list of coordinates and find the start and end points
    where the y-value is not zero. Raise exception if no points found.'''

    start = -1
    end = -1
    for i in range(len(coords)):
        if coords[i][1] != 0:
            if start == -1:
                start = i
            end = i

    if start == -1:
        raise Exception("No unique k-mers found. Please check input or increase k-mer size range.")
        exit(1)

    return start, end


def line_y(p1, p2, x):
    '''Given points p1 and p2, define a straight line function between them and
    return the coresponding y-value given the x-value. Used in finding the knee
    point.'''

    # Line equation: y - p1.y = (p2.y - p1.y) / (p2.x - p1.x) * (x - p1.x)
    y = float(p2[1] - p1[1]) / float(p2[0] - p1[0])
    y = y * (x - p1[0])
    y = y + p1[1]

    return y


def interpret_stats(stats_string):
    '''Given an output string from jellyfish stats, decipher it and return the
    useful information.'''
    unique = re.search(r'Unique:\s+(?P<unique>[\d]+)', stats_string).group('unique')
    distinct = re.search(r'Distinct:\s+(?P<distinct>[\d]+)', stats_string).group('distinct')
    total = re.search(r'Total:\s+(?P<total>[\d]+)', stats_string).group('total')

    ratio = 0
    if int(distinct) != 0:
        ratio = float(unique) / float(distinct)

    log_string = '\t' + unique + '\t' + distinct + '\t' + total + '\n'

    return ratio, log_string


def clean_up_files(file_list, is_verbose):
    '''Goes through a list of file paths and remove them.'''

    for filename in file_list:
        if is_verbose:
            print "Removing file:", filename

        try:
            os.remove(filename)
            logging.info(filename + " ... removed.\n")
        except OSError, e:
            msg = "Failed to clean file: \"" + filename + "\" " + e
            logging.critical(msg)
            print >> sys.stderr, msg


def check_spawning_site(out_dir):
    '''Given the output directory path, creates output directory if it does not
    exist.'''
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)


def spawn_logger(out_dir, is_verbose):
    '''Given the output directory path, sets up a logger.'''
    time_str = str(time.time())
    log_filename = out_dir + '/spawn_' + time_str + '.log'

    if is_verbose:
        logging.basicConfig(filename=log_filename, format='%(asctime)s %(levelname)s: %(message)s', datefmt='%Y-%b-%d %H:%M:%S', level=logging.DEBUG)
    else:
        logging.basicConfig(filename=log_filename, format='%(asctime)s %(message)s', level=logging.INFO)


def check_params(parser, args, config):
    if args.ncpus <= 0 or args.kmer_size <= 0:
        print "\nNumber of CPUs and k-mer size must be positive integers"
    elif config.getint('JELLYFISH', 'MIN_KMER_SIZE') <= 0 or config.getint('JELLYFISH', 'MAX_KMER_SIZE') <= 0:
        print "\nK-mer range must be positive integers"
    else:
        return True

    return False


if __name__ == '__main__':

    # set input options
    descriptions = "PAV Pipeline: spawn jellyfish k-mer counter instances."
    parser = argparse.ArgumentParser(description=descriptions)
    parser.add_argument("-s", dest="sample", nargs='+', required=True, help="input paired reads fasta files")
    parser.add_argument("-o", "--out_dir", dest="out_dir", help="output directory")
    parser.add_argument("-t", "--tag", dest="tag", help="jellyfish output name tag", default="jellies")
    parser.add_argument("-c", "--config_file", dest="config", required=True, help="input config file")
    parser.add_argument("-k", "--kmer_size", dest="kmer_size", help="input the k-mer size to compute", type=int, default=17)
    parser.add_argument("-n", "--ncpus", dest="ncpus", help="input the number of CPUs to use [default=%(default)s]", type=int, default=1)
    parser.add_argument("-m", "--mode", dest="mode", help="jellyfish dump or stats [default: %(default)s]", choices=["dump", "stats"], default="dump")
    parser.add_argument("-v", "--verbose", dest="is_verbose", action="store_true", help="be verbose [default=%(default)s]", default=False)

    args = parser.parse_args()

    config = ConfigParser.ConfigParser()
    config.read(args.config)

    k_min = config.getint('JELLYFISH', 'MIN_KMER_SIZE')
    k_max = config.getint('JELLYFISH', 'MAX_KMER_SIZE')

    if not check_params(parser, args, config):
        raise IOError("Incorrect arguments. Quitting.")

    check_spawning_site(args.out_dir)
    spawn_logger(args.out_dir, args.is_verbose)
    logging.info('Begin jellyfish_spawn ...\n\tMode is: %s\n\tInput sample file(s): %s\n\tConfig file: %s\n\tNumber of CPUs: %s\n', args.mode, ' '.join(args.sample), args.config, str(args.ncpus))

    if args.mode == "dump":
        jellyfish_polyp(config, args.kmer_size, args.sample, args.out_dir, args.tag, args.ncpus, args.mode, args.is_verbose)
    elif args.mode == "stats":
        jellyfish_bloom(config, args.sample, args.out_dir, args.tag, args.ncpus, args.mode, k_min, k_max, args.is_verbose)

    logging.info('Finished spawning')
