#!/usr/bin/env python
################################################################################
# PAV Pipeline: split a file into sub-files based on the first n bases.
# Note: jellyfish count outputs expected. Each line should be in the form of:
# >K-mer occurrence
#
# Hong Ching Lee     30 June 2011
################################################################################

# import libraries
import argparse # parse input options
import math # mathematical operations
import re # regular expressions


def make_n_mers(n, is_verbose=False):
    '''This method returns a dict mapping the n_mers to a number.'''

    nucleotides = ['A', 'C', 'G', 'T', 'N'] # base nucleotides
    num_mers = int(math.pow(len(nucleotides), n)) # the number of n_mers required
    n_mers = [""] * num_mers # the pre-initialised array to store the n_mers
    n_mers_map = {} # the returned dictionary storing the n_mers as key

    if is_verbose:
        print "Initialised n_mers array:", n_mers

    # go through each loop to add a nucleotide at the end of each n_mers element
    # essentially goes through the n_mers list n number of times, each time
    # appending a base to each element.
    n_times = num_mers
    for i in range(n):
        # number of times to put each nucleotide before moving to the next.
        # so if n = 2, in the first round it puts 4 A's, 4 G's etc.; second
        # round 1 A's... To make A, A, A, A, C, C, C, C, G, G, G, G, T, T, T, T
        # then AA, AC, AG, AT, CA, CC, CG, CT, ...
        n_times = n_times / len(nucleotides)
        # number of rounds to go through the nucleotides list
        n_rounds = num_mers / len(nucleotides) / n_times

        # main inner loop for actually appending nucleotide bases to each n_mers
        mer_i = 0
        for round_i in range(n_rounds):
            for nuc_i in range(len(nucleotides)):
                for times_i in range(n_times):
                    n_mers[mer_i] = n_mers[mer_i] + nucleotides[nuc_i]
                    mer_i += 1

    if is_verbose:
        print "Split by these n_mers:"
        for i in range(num_mers):
            print i, ":", n_mers[i]

    for i in range(num_mers):
        n_mers_map[n_mers[i]] = i

    return num_mers, n_mers_map


def split_file(in_file_name, n, num_mers, n_mers_map, is_verbose=False):
    '''This method does the actual splitting of the input file into exactly
    num_mers number of output files.'''

    handles = [] # create the output file handles
    for i in range(num_mers):
        handles.append(open(in_file_name + "." + str(i), "w"))

    # go through the input file
    with open(in_file_name, "rU") as in_f:
        for line in in_f:
            splitted_line = line.split()
            occ = int(splitted_line[1])

            # get the n-mer, or first n bases of the k-mer string
            n_mer = splitted_line[0][0:n]
            if is_verbose:
                print "k-mer is:", splitted_line[0], "and n-mer is:", n_mer

            handles[n_mers_map[n_mer]].write(line)

    in_f.close()

    for i in range(num_mers):
        handles[i].close()


def check_param(parser, args):
    '''This method checks the input parameters.'''

    if (args.in_file_name is None):
        print("===Please specify an input file.===")
        parser.print_help()
    elif (args.n_bases < 0):
        print("===Number of bases to split needs to be 0 or above. -n 0, no split. -n 1 means split the input 4 times using the first base, -n 2 splits 16 times using first 2 bases===")
        parser.print_help()
    else:
        return True

    return False


if __name__ == '__main__':

    # set input options
    usage = "%(prog)s -i <input_file> -n <n_bases>\n\n"
    descriptions = "For PAV Pipeline: split a file into sub-files based on the first n bases of each k-mer.\nNote: jellyfish count outputs expected. Each line should be in the form of:\n>K-mer occurrence"
    parser = argparse.ArgumentParser(usage=usage, description=descriptions)
    parser.add_argument("-i", "--in", dest="in_file_name", required=True, help="input file name")
    parser.add_argument("-n", dest="n_bases", help="Number of bases at the beginning of the mer to split the input. [default=%(default)s]", type=int, default=2)
    parser.add_argument("-v", "--verbose", dest="is_verbose", action="store_true", help="be verbose [default=%(default)s]", default=False)

    args = parser.parse_args()

    if check_param(parser, args):
        # create the n_mers which will be used to match the starting n-mers of
        # each line. Also get the number of mers there are
        num_mers, n_mers_map = make_n_mers(args.n_bases, args.is_verbose)

        # split the input file and output to different files using the first n
        # bases of the input k-mers
        split_file(args.in_file_name, args.n_bases, num_mers, n_mers_map, args.is_verbose)

